/*
 * tmos - sh.c
 * 	version 0.0.0
 * 
 * Copyright (C) 2009  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/console.h>
#include <tmos/serial.h>
#include <tmos/stdlib.h>
#include <tmos/string.h>
#include <tmos/port.h>
#include <tmos/mutex.h>
#include <limits.h>

#define CMD_LEN_MAX	24
#define SH_PROMPT	"sh# "
#define MAX_JOBS	CONFIG_SH_JOBS_MAX

#if CONFIG_PORT
static port_t port;

static DECLARE_MUTEX(put_lock);

/**
 * getc is not thread-safe!
 */
static char getc(void) __reentrant
{
	static char __user t;

	port_read(port, &t, 1);
	return t;
}

static void __putc(char c) __reentrant
{
	static char __user t;

	t = c;
	sys_port_write(port, &t, 1);
}

static void putc(char c) __reentrant
{
	kernel_enter();
	mutex_lock(put_lock);
	__putc(c);
	mutex_unlock(put_lock);
	kernel_exit();
}

static void puts(const char *s) __reentrant
{
	while (*s)
		putc(*s++);
}

#define write(buf, len)	port_write(port, (buf), (len))

#elif CONFIG_SERIAL
#define getc()		ser_rxc()
#define putc(c)		ser_txc((c))
#define puts(s)		ser_txs((s))
#define write(buf, len)	ser_tx_user((buf), (len))
#else
#error "I/O not supported"
#endif


#if CONFIG_SH_JOBS
#include <tmos/sched.h>

static char __kernel null;
static struct context __kernel sh_context;
static struct context __kernel job[MAX_JOBS];
// static unsigned char __user jobs = 0;
static unsigned char __user __cmd_lock = UCHAR_MAX;
static unsigned char __user cmd_len = 0;

extern struct context __kernel *current;

#define cmd_lock(x) \
	do { \
		smp_wmb(); \
		__cmd_lock = x; \
	} while (0)

#define cmd_unlock() cmd_lock(UCHAR_MAX)

#define cmd_wait(x) \
	do { \
		while (x >= __cmd_lock) \
			schedule(); \
	} while (0)

#define start_jobs() \
	do { \
		context_init(&sh_context); \
		sh_context.next = &sh_context; \
		current = &sh_context; \
	} while (0)

BEGIN_TASK_ARRAY
	TASK(1, "sh", &sh_context),
	TASK(2, "job0", &job[0]),
END_TASK_ARRAY

#else
#define start_jobs() do { } while (0)
#define cmd_lock(x) do { } while (0)
#define cmd_unlock(x) do { } while (0)
#define cmd_wait(x) do { } while (0)
#endif

static unsigned char __user cmd[CMD_LEN_MAX + 1];
static unsigned char __user cmd_i = 0;

static char d_to_char(char x) __user_preempt
{
	return ((x >= 10) ? 'a' - 10 : '0') + x;
}

static void print_int(unsigned int val) REENTRANT
{
	unsigned char x;

	if (val < 10)
		goto print_1;
	if (val < 100)
		goto print_10;
	if (val < 1000)
		goto print_100;
	if (val < 10000)
		goto print_1000;
	
	x = val / 10000;
	putc(d_to_char(x));
	val -= 10000 * x;
	
print_1000:
	x = val / 1000;
	putc(d_to_char(x));
	val -= 1000 * x;
	
print_100:
	x = val / 100;
	putc(d_to_char(x));
	val -= 100 * x;

print_10:
	x = val / 10;
	putc(d_to_char(x));
	val -= 10 * x;
print_1:
	x = val;
	putc(d_to_char(x));
}

static void print_int_n(unsigned int val, unsigned char n) REENTRANT
{
	unsigned char x;
	
	if (val >= 10)
		n--;
	if (val >= 100)
		n--;
	if (val >= 1000)
		n--;
	if (val >= 10000)
		n--;

	for (; n > 1; n--)
		putc(' ');

	if (val < 10)
		goto print_1;
	if (val < 100)
		goto print_10;
	if (val < 1000)
		goto print_100;
	if (val < 10000)
		goto print_1000;
	
	x = val / 10000;
	putc(d_to_char(x));
	val -= 10000 * x;
	
print_1000:
	x = val / 1000;
	putc(d_to_char(x));
	val -= 1000 * x;
	
print_100:
	x = val / 100;
	putc(d_to_char(x));
	val -= 100 * x;

print_10:
	x = val / 10;
	putc(d_to_char(x));
	val -= 10 * x;
print_1:
	x = val;
	putc(d_to_char(x));
}

static void print_hex(unsigned char val) REENTRANT
{
	putc(d_to_char((val >> 4) & 0xf));
	putc(d_to_char((val >> 0) & 0xf));
}

static void print_whex(unsigned int val) REENTRANT
{
	putc(d_to_char(val >> 12));
	putc(d_to_char((val >> 8) & 0xf));
	putc(d_to_char((val >> 4) & 0xf));
	putc(d_to_char((val >> 0) & 0xf));
}
			
static void show_prompt(void) REENTRANT
{
	puts(SH_PROMPT);
}

#if CONFIG_TASK_TIMES && CONFIG_SH_JOBS
static void sh_ps(void)
{
	DECLARE_TASK_POINTER(task);

	puts("  PID TTY      TIME   SP  WCHAN CMD\n");
	for_each_task(task) {
		print_int_n(task->pid, 5);
		puts(" ttyS0   ");
		print_int_n(jiffies_ms(task->context->cpu_time), 5);
		puts("   ");
		print_int_n(task->context->sp, 3);
		puts(" ");
		if (task->context == current) {
			puts("RUN ");
		} else if (task->context->sp == 0) {
			puts("    ");
		} else {
			print_whex(GET_WCHAN(task->context));
		}
		puts(" ");
		puts(task->comm);
		putc('\n');
	}
}

static void sh_top(void) REENTRANT
{
	jiffies_t t[2];
	unsigned int c;
	DECLARE_TASK_POINTER(task);

	for (;;) {
		c = 0;
		for_each_task(task)
			t[c++] = task->context->cpu_time;
#if 1
#if CONFIG_PORT
		{
			static unsigned char __user c;
			if (port_read_t(port, &c, 1, 1000) == 1)
				break;
		}
#else
		if (ser_rx_kernel_t(&null, 1, 1000) == 1)
			break;
#endif
#else
		jsleep(1000);
#endif
		
		puts("  PID TTY      CPU  CMD\n");
		c = 0;
		for_each_task(task) {
			t[c] = task->context->cpu_time - t[c];

			print_int_n(task->pid, 5);
			puts(" ttyS0    ");
			print_int_n(t[c] / 10, 3);
			puts("  ");
			puts(task->comm);
			putc('\n');
			c++;
		}
	}
}
#endif

#if CONFIG_PORT
static void do_cat(unsigned char __user *cmd)
{
	port_t port;
	msg_ssize_t ret;
	char __user buf[16];

	port = port_open(cmd, O_RDONLY);
	if (port == (port_t) -1) {
		puts("cat: ");
		puts(cmd);
		puts(": No such file or directory\n");
		return;
	}
	for (;;) { 
		ret = port_read(port, buf, 16);
		if (ret == 0) {
			puts("zero\n");
			break;
		}
		if (ret < 0) {
			puts("cat: ");
			puts(cmd);
			puts(": read error\n");
			break;
		}
		write(buf, ret);
	}
}

static void do_hexdump(unsigned char __user *cmd)
{
	port_t port;
	off_t off = 0;
	msg_ssize_t ret;
	char __user buf[16];
	unsigned char i;

	port = port_open(cmd, O_RDONLY);
	if (port == (port_t) -1) {
		puts("hexdump: ");
		puts(cmd);
		puts(": No such file or directory\n");
		return;
	}
	for (;;) { 
		ret = port_pread(port, buf, 16, off);
		if (ret == 0)
			break;
		if (ret < 0) {
			puts("hexdump: ");
			puts(cmd);
			puts(": read error\n");
			break;
		}
		print_whex(off);
		puts(": ");
		off += ret;
		for (i = 0; i < ret; i++) {
			print_hex(buf[i]);
			putc(' ');
			if (i == 7)
				putc(' ');
		}
		puts("\n");
	}
}
#endif

static void __process_command(unsigned char offset, unsigned char cmd_len,
		unsigned char child)
	REENTRANT
{
	unsigned char __user *c = cmd + offset;
	unsigned char sep;

	for (sep = offset; sep < cmd_len; sep++)
		if (isspace(cmd[sep])) {
			cmd[sep] = '\0';
			break;
		}

	if (!strcmp(c, "echo")) {
		if (sep < cmd_len)
			puts(cmd + sep + 1);
		putc('\n');
#if CONFIG_PORT
	} else if (!strcmp(c, "cat")) {
		if (sep < cmd_len)
			do_cat(cmd + sep + 1);
		putc('\n');
	} else if (!strcmp(c, "hexdump")) {
		if (sep < cmd_len)
			do_hexdump(cmd + sep + 1);
		putc('\n');
#endif
	} else if (!strcmp(c, "dump_stack")) {
		dump_stack();
	} else if (!strcmp(c, "time")) {
		jiffies_t t = get_jiffies();
		if (child)
			cmd_lock(sep);
		__process_command(sep + 1, cmd_len, child);
		t = get_jiffies() - t;
		puts("real: ");
		print_int(jiffies_ms(t));
		puts("ms\n");
	} else if (!strcmp(c, "sleep")) {
	       	if (sep < cmd_len) {
			unsigned int sec = atoi(cmd + sep + 1);
			if (child)
				cmd_lock(CMD_LEN_MAX + 1);
			sleep(sec);
		}
#if CONFIG_TASK_TIMES && CONFIG_SH_JOBS
	} else if (!strcmp(c, "ps")) {
		if (child)
			cmd_lock(CMD_LEN_MAX + 1);
		sh_ps();
	} else if (!strcmp(c, "top")) {
		if (child)
			cmd_lock(CMD_LEN_MAX + 1);
		sh_top();
#endif
	} else {
		puts("-sh: ");
		puts(c);
		puts(": command not found\n");
	}
}

static void process_command(void) REENTRANT
{
	if (!cmd_i)
		return;

	cmd_wait(cmd_i);
	cmd[cmd_i] = '\0';

#if CONFIG_SH_JOBS
	if (cmd[cmd_i - 1] == '&') {
		cmd_wait(UCHAR_MAX - 1);
		cmd_len = cmd_i - 1;

		context_init(&job[0]);

		/*
		 * thread_create currently must be called with preemption
		 * disabled.
		 */
		kernel_enter();
		preempt_disable();
		if (thread_create(&job[0], (unsigned char)0xc0)) {
			__process_command(0, cmd_len, 1);
			sh_context.next = &sh_context;
			cmd_unlock();
			schedule();
		}
		preempt_enable();
		kernel_exit();
		job[0].next = sh_context.next;
		sh_context.next = &job[0];
	} else
#endif
	{
		__process_command(0, cmd_i, 0);
	}
	cmd_i = 0;
}

void app_main(void) REENTRANT
{
#if CONFIG_PORT
	/* FIXME: we should use user function */
	port = sys_port_open(CONFIG_SH_DEVICE, O_RDWR);
	if (port == (port_t) -1)
		BUG();
#endif
	puts("tmos sh 0.0.0\n");
	start_jobs();
	show_prompt();
	for (;;) {
		char c;

		c = getc();
		if (c == '\b' || c == 127) {
			if (cmd_i) {
				puts("\b \b");
				cmd_i--;
			}
		} else if (c == '\n' || c == '\r') {
			putc(c);
			process_command();
			show_prompt();
		} else if (cmd_i < CMD_LEN_MAX) {
			putc(c);
			cmd_wait(cmd_i);
			cmd[cmd_i] = c;
			cmd_i++;
		}
	}
}
