/*
 * tmos - aduc841.c
 * 	version 0.0.0
 * 
 * Copyright (C) 2009  Marek Piwonski <marek@microster.pl>
 * Copyright (C) 2009, 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/adc.h>
#include <asm/aduc841.h>
#include <asm/irqs.h>

// #define ADC_DEBUG

#define ADC_TEMP		0x8
#define ADC_DAC0		0x9
#define ADC_DAC1		0xa
#define ADC_AGND		0xb
#define ADC_VREF		0xc
#define ADC_STOP		0xf

#define ADCCON1_MD1		0x80
#define ADCCON1_EXT_REF		0x40
#define ADCCON1_CK_32		0x00
#define ADCCON1_CK_4		0x10
#define ADCCON1_CK_8		0x20
#define ADCCON1_CK_2		0x30
#define ADCCON1_AQ_1		0x00
#define ADCCON1_AQ_2		0x04
#define ADCCON1_AQ_3		0x08
#define ADCCON1_AQ_4		0x0c
#define ADCCON1_T2C		0x02
#define ADCCON1_EXC		0x01

#define ADCCON2_ADCI	0x80
#define ADCCON2_DMA	0x40
#define ADCCON2_CCONV	0x20
#define ADCCON2_SCONV	0x10
#define ADCCON2_CH(x)	((x) & 0xf)

#define ADCCON3_BUSY	0x80
#define ADCCON3_AVG1	0x10
#define ADCCON3_AVG15	0x00
#define ADCCON3_AVG31	0x20
#define ADCCON3_AVG63	0x30
#define ADCCON3_CALIB	0x04
#define ADCCON3_TYPICAL	0x02
#define ADCCON3_SCAL	0x01

#define ADCCON3_TYPICAL_GAIN	ADCCON3_TYPICAL
#define ADCCON3_TYPICAL_OFFSET	0

#define RELOAD_VALUE \
	(uint16_t)(65536 - (CONFIG_OSCILLATOR / CONFIG_ADC_ADUC841_FREQ))

#ifdef CONFIG_ADC_ADUC841_USE_XDATA
#define bdata __xdata
#else
#define bdata __data
#endif

#define F_N		CONFIG_ADC_ADUC841_CHANNEL_COUNT

static uint8_t nr_ch;

#ifdef CONFIG_ADC_SYNC
static uint8_t nr_seq;

#define adc_seq_inc()	do { nr_seq++; } while (0)
	
/**
 * adc_sync - wait until channel ch is ready
 * @ch: channel
 */
unsigned char adc_sync(unsigned char seq) __reentrant
{
	while (nr_seq == seq)
		cpu_relax();
	smp_rmb();
	return nr_seq;
}

/**
 * adc_wait_ch - wait until channel ch is ready
 * @ch: channel
 */
void adc_wait_ch(unsigned char ch) __reentrant
{
	while (nr_ch == ch)
		cpu_relax();
	smp_rmb();
}
#else
#define adc_seq_inc()	do { } while (0)
#endif

#if CONFIG_ADC_ADUC841_FILTER_2ND

#define NEW_FILTER		1

#define f_data bdata

/*
 * new filter interface:
 * 
 * The filter code should export:
 * 	f_process(CH, VALUE) - called by ADC driver on every sample
 * 	f_get(CH) - called by ADC driver to read channel CH
 */
#define f_ch	nr_ch

#define FILTER_CODE static void f_process(int16_t x) __using ADUC841_ADC_BANK

#if CONFIG_ADC_ADUC841_FILTER_2ND
#include "filter-2nd.h"
#endif

#define update_value(ch, val) f_process(val)

int16_t adc_get(unsigned int id)
{
	if (id >= CONFIG_ADC_ADUC841_CHANNEL_COUNT)
		return 0;
	return f_get(id);
}
#else

static uint16_t bdata advalue[CONFIG_ADC_ADUC841_CHANNEL_COUNT];

int16_t adc_get(unsigned int id)
{
	if (id >= CONFIG_ADC_ADUC841_CHANNEL_COUNT)
		return 0;
	return advalue[id] - 32768;
}


#ifdef CONFIG_ADC_ADUC841_NO_FILTER
#define update_value(ch, val) \
	do { \
		advalue[ch] = val << 4; \
	} while (0)

#elif CONFIG_ADC_ADUC841_FILTER_6
/*
 * filter for 6x oversampling
 */
static uint16_t bdata y1[CONFIG_ADC_ADUC841_CHANNEL_COUNT];
static uint16_t bdata y2[CONFIG_ADC_ADUC841_CHANNEL_COUNT];
static uint16_t bdata xr[CONFIG_ADC_ADUC841_CHANNEL_COUNT][8];
static unsigned char index;

#define update_value(ch, val) \
	do { \
		uint16_t tmp; \
		tmp = y1[ch] + val + xr[ch][index] + xr[ch][(index + 4) & 7]; \
		tmp = (tmp + 2) >> 2; \
		y1[ch] = tmp; \
		tmp += (y2[ch] << 1) + y2[ch]; \
		tmp = (tmp + 2) >> 2; \
		y2[ch] = tmp; \
		tmp = advalue[ch]; \
		tmp = (tmp + 2) >> 2; \
		tmp += (tmp << 1); \
		tmp += y2[ch]; \
		advalue[ch] = tmp; \
		xr[ch][index] = val; \
	} while (0)
		
#define update_index()	do { index++; } while (0)

#elif CONFIG_ADC_ADUC841_FILTER_10
/*
 *            1
 * y = ---------------- u
 *       (4 - 3 z^-1)
 */
#define update_value(ch, val) \
	do { \
		advalue[ch] += (val << 2) - (advalue[ch]) / 4; \
	} while (0)

#elif CONFIG_ADC_ADUC841_FILTER_40

static uint16_t bdata y1[CONFIG_ADC_ADUC841_CHANNEL_COUNT];

/*
 *              1
 * y = ------------------ u
 *       (8 - 7 z^-1)^2
 */
#define update_value(ch, val) \
	do { \
		y1[ch] += (val << 1) - y1[ch] / 8; \
		advalue[ch] += y1[ch] / 8 - advalue[ch] / 8; \
	} while (0)

#elif CONFIG_ADC_ADUC841_FILTER_80

static uint16_t bdata y1[CONFIG_ADC_ADUC841_CHANNEL_COUNT];

/*
 *              1
 * y = ------------------ u
 *       (16 - 15 z^-1)^2
 */
#define update_value(ch, val) \
	do { \
		y1[ch] += val - y1[ch] / 16; \
		advalue[ch] += y1[ch] / 16 - advalue[ch] / 16; \
	} while (0)

#elif CONFIG_ADC_ADUC841_FILTER_50
/*
 *           1
 * y = -------------- u
 *       16 - 15 z^-1
 */
#define update_value(ch, val) \
	do { \
		advalue[ch] += val - (advalue[ch]) / 16; \
	} while (0)
#else
#error "unimplemented filter"
#endif

#ifndef update_index
#define update_index()	do { } while (0)
#endif
#endif

#define start_conv_ch() \
	do { \
		ADCCON2 = nr_ch | ADCCON2_SCONV; \
	} while (0)

#if 0
static uint16_t adc_read(void) __using ADUC841_ADC_BANK __naked
{
	__asm
; high byte
		mov	a, #0x0F
		anl	a, _ADCDATAH
		add	a, #0xf8
		mov	dph, a
; low byte
		mov	a, _ADCDATAL
		mov	dpl,a
		ret
	__endasm;	
}
#else
static uint16_t adc_read(void) __using ADUC841_ADC_BANK
{
	uint8_t high = ADCDATAH & 0x0f;
	uint8_t low = ADCDATAL;
	uint16_t x = (high << 8) | low;

#if NEW_FILTER
	return (int16_t) x - 2048;
#else
	return x;
#endif
}
#endif

void aduc_timer_irq(void) __interrupt ADUC_TIMER_INTERRUPT
		__using ADUC_TIMER_BANK
{
#if NEW_FILTER
	uint16_t x;
#else
	int16_t x;
#endif
#ifdef ADC_DEBUG
	static uint16_t c;
	tsc_t tt;

	tt = tsc_read_low();
#endif
	if (!ADCI)
		goto out;
	x = adc_read();
	update_value(nr_ch, x);

/*	smp_wmb(); */
	nr_ch++;
#if CONFIG_ADC_ADUC841_CHANNEL_COUNT == 8
	nr_ch &= 0x07;
#elif CONFIG_ADC_ADUC841_CHANNEL_COUNT == 4
	nr_ch &= 0x03;
#elif CONFIG_ADC_ADUC841_CHANNEL_COUNT == 2
	nr_ch &= 0x01;
#else
	if (nr_ch >= CONFIG_ADC_ADUC841_CHANNEL_COUNT)
		nr_ch = 0;
#endif
/*	smp_wmb(); */
	if (!nr_ch)
		adc_seq_inc();

	start_conv_ch();

#ifdef ADC_DEBUG
	tt = tsc_read_low() - tt;
	f_y[0] = tt / 2;
	c++;
	f_y[1] = c / 2;
	f_y[2] = __get_jiffies() / 2;
#endif

out:
	TF2 = 0;	/* clear interrupt bit */
}

void aduc_timer_init(void)
{
#if ADUC_TIMER2
	TR2 = 0;			//stop the timer2
	TF2 = 0;			//clear pending interrupt
	RCLK = 0;			//serial don't use this timer
	TCLK = 0;			//serial don't use this timer
	CAP2 = 0;			//autoreload mode
	EXEN2 = 0;			//normal reload
	CNT2 = 0;			//use on-chip clock source
	TL2 = RELOAD_VALUE & 0xff;	//initial L value
	TH2 = RELOAD_VALUE >> 8;	//initial H value
	RCAP2L = RELOAD_VALUE & 0xff;	//reload L value
	RCAP2H = RELOAD_VALUE >> 8;	//reload H value
#if CONFIG_ADC_ADUC841_TIMER_HIGHPRIO
	PT2 = 1;
#endif
	ET2 = 1;			//enable timer2 interrupt
	TR2 = 1;			//strat the timer
#elif ADUC_TIMER1
	TMOD = (TMOD & ~0xf0) | 0x10;
	TL1 = RELOAD_VALUE & 0xff;
	TH1 = RELOAD_VALUE >> 8;
#if CONFIG_ADC_ADUC841_TIMER_HIGHPRIO
	PT1 = 1;
#endif
	TF1 = 1;
	ET1 = 1;
#else
#error "unsupported ADC timer"
#endif
}

static inline void adc_wait(void)
{
	while (ADCCON3 & ADCCON3_BUSY);
}

static void ad_ofs_cal(void)
{
	ADCCON2 = ADCCON2_SCONV	| ADCCON2_CH(ADC_AGND);
	ADCCON3 = ADCCON3_AVG63 | ADCCON3_CALIB | ADCCON3_TYPICAL_OFFSET
		| ADCCON3_SCAL;
	adc_wait();
}

static void ad_gain_cal(void)
{
	ADCCON2 = ADCCON2_SCONV	| ADCCON2_CH(ADC_VREF);
	ADCCON3 = ADCCON3_AVG63 | ADCCON3_CALIB | ADCCON3_TYPICAL_GAIN
		| ADCCON3_SCAL;
	adc_wait();
}

void adc_init(void)
{
/*	ADCLK = OSC/2, Aquisition time = 4 ADCLK */
	ADCCON1 = ADCCON1_MD1 | ADCCON1_CK_2 | ADCCON1_AQ_4;
	ADCCON2 = 0;
	ADCCON3 = ADCCON3_AVG1;
	EADC = 0;
	ad_ofs_cal();
	ad_gain_cal();
	aduc_timer_init();
}

device_initcall(adc_init)
