/*
 * tmos - ADUC EEPROM data memory
 * Copyright (C) 2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <tmos/kernel.h>
#include <tmos/port.h>
#include <tmos/errno.h>
#include <asm/aduc841.h>

#define EE_PAGE_SIZE	4
#define EE_PAGE_MASK	(EE_PAGE_SIZE - 1)
#define EE_PAGES	1024

#define EE_READ		1
#define EE_WRITE	2
#define EE_VERIFY	4
#define EE_ERASE	5
#define EE_ERASE_ALL	6
#define EE_READBYTE	0x81
#define EE_WRITEBYTE	0x82
#define EE_EXULOAD	0x0f
#define EE_ULOAD	0xf0

#if (EE_PAGES * EE_PAGE_SIZE) > OFF_MAX
#define EE_SIZE	OFF_MAX
#else 
#define EE_SIZE		(EE_PAGES * EE_PAGE_SIZE)
#endif

#define ee_read_page()		do { ECON = EE_READ; } while (0)
#define ee_erase_page()		do { ECON = EE_ERASE; } while (0)
#define ee_write_page()		do { ECON = EE_WRITE; } while (0)
#define ee_verify_page()	do { ECON = EE_VERIFY; } while (0)
#define ee_verify_ok()		(ECON == 0)

static msg_ssize_t ee_write(char __user *buf, msg_size_t len, off_t off) 
	__reentrant
{
	msg_ssize_t ret = 0;

	while (len > 0) {
		unsigned char d;
		unsigned char chunk;
		unsigned char i;
		EADRH = off / (EE_PAGE_SIZE * 256);
		EADRL = off / EE_PAGE_SIZE;

		chunk = EE_PAGE_SIZE - (off & EE_PAGE_MASK);
		if (chunk > len)
			chunk = len;

		/*
		 * READ-MODIFY-WRITE for partial page writes
		 */
		if (chunk != EE_PAGE_SIZE)
			ee_read_page();

		for (i = 0; i < chunk; i++) {
			get_user(d, buf);

			switch (off & EE_PAGE_MASK) {
			case 0:
				EDATA1 = d;
				break;
			case 1:
				EDATA2 = d;
				break;
			case 2:
				EDATA3 = d;
				break;
			case 3:
				EDATA4 = d;
				break;
			} 
			off++;
			buf++;
		}
		ee_erase_page();
		ee_write_page();
		ee_verify_page();
		if (!ee_verify_ok()) {
			if (!ret)
				ret = -EIO;
			break;
		}
		ret += chunk;
		len -= chunk;
	}
	return ret;
}

static msg_ssize_t ee_read(char __user *buf, msg_size_t len, off_t off) 
	__reentrant
{
	msg_ssize_t ret = 0;

	while (len > 0) {
		unsigned char d = 0;
		unsigned char chunk;
		unsigned char i;

		EADRH = off / (EE_PAGE_SIZE * 256);
		EADRL = off / EE_PAGE_SIZE;

		chunk = EE_PAGE_SIZE - (off & EE_PAGE_MASK);
		if (chunk > len)
			chunk = len;

		ee_read_page();

		for (i = 0; i < chunk; i++) {
			switch (off & EE_PAGE_MASK) {
			case 0:
				d = EDATA1;
				break;
			case 1:
				d = EDATA2;
				break;
			case 2:
				d = EDATA3;
				break;
			case 3:
				d = EDATA4;
				break;
			} 
			put_user(d, buf);
			off++;
			buf++;
		}
		ret += chunk;
		len -= chunk;
	}
	return ret;
	
}
	
msg_ssize_t dev_eedata_xmit(port_t port, char __user *buf,
		msg_size_t len, msg_size_t max_resp_len, msg_flags_t flags,
		off_t off)
	__reentrant
{
	msg_ssize_t ret = 0;
		
	if (off >= EE_SIZE)
		return 0;

	if (len) {
		if (len > EE_SIZE - off)
			len = EE_SIZE - off;
		ret = ee_write(buf, len, off);
	}

	if (max_resp_len) {
		if (max_resp_len > EE_SIZE - off)
			max_resp_len = EE_SIZE - off;
		ret = ee_read(buf, max_resp_len, off);
	}
	return ret;
}
