/*
 * tmos - system.h
 *
 * Copyright (C) 2009  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __ASM_SYSTEM_H
#define __ASM_SYSTEM_H

#include <asm/barrier.h>
#include <asm/processor.h>
#include <asm/8052.h>
#include <tmos/config.h>
#include <string.h>

/*
 * We should use the same bank for ALL interrupts with the same priority.
 */
#define HIGH_PRIORITY_BANK	2
#define LOW_PRIORITY_BANK	3

typedef unsigned char cpu_flags_t;

#define local_save_flags(x)	do { (x) = EA; } while (0)
#define local_irq_restore(x)	do { EA = (x); } while (0)
#define local_irq_disable()	do { EA = 0; } while (0)
#define local_irq_enable()	do { EA = 1; } while (0)
#define local_irq_save(x)						\
	do {								\
		x = EA;							\
		EA = 0;							\
	} while (0)

#if CONFIG_USER_PREEMPT && CONFIG_KERNEL_PREEMPT
extern char preempt_count;

#define preempt_disable() \
	do { \
		preempt_count++; \
		barrier(); \
	} while (0)

#define preempt_enable() \
	do { \
		barrier(); \
		preempt_count--; \
	} while (0)

#define preempt_enabled() (!preempt_count)
#define preempt_enabled_flags(flags) (!preempt_count)

#define kernel_enter()	do { } while (0)
#define kernel_exit()	do { } while (0)

#elif CONFIG_USER_PREEMPT

#define preempt_disable()	do { } while (0)
#define preempt_enable()	do { } while (0)

#define kernel_enter() \
	do { \
		F1 = 1; \
		barrier(); \
	} while (0)

#define kernel_exit() \
	do { \
		barrier(); \
		F1 = 0; \
	} while (0)

#define preempt_enabled() (!F1)
#define preempt_enabled_flags(flags) (!(flags & 0x02))

#else

#define preempt_disable()		do { } while (0)
#define preempt_enable()		do { } while (0)
#define preempt_enabled()		(0)
#define preempt_enabled_flags(flags)	(0)

#define kernel_enter()	do { } while (0)
#define kernel_exit()	do { } while (0)
#endif

#if CONFIG_KERNEL_XDATA
#define __kernel __xdata
typedef uint16_t kernel_size_t;
#define KERNEL_TASK_SIZE	65536
#elif CONFIG_KERNEL_DATA
#define __kernel __data
typedef uint8_t kernel_size_t;
#define KERNEL_TASK_SIZE	256
#else
#define __kernel
typedef size_t kernel_size_t;
#define KERNEL_TASK_SIZE	(1 << 24UL)
#endif

#if CONFIG_USER_XDATA
#define __user __xdata
typedef uint16_t user_size_t;
#elif CONFIG_USER_DATA
#define __user __data
typedef uint8_t user_size_t;
#else
#define __user
typedef size_t user_size_t;
#endif

#if CONFIG_USER_PREEMPT
#define __user_preempt __reentrant
#else
#define __user_preempt
#endif

#if CONFIG_KERNEL_PREEMPT
#define __kernel_preempt __reentrant
#else
#define __kernel_preempt
#endif

#define copy_to_user(dst, src, len)	(memcpy(dst, src, len), 0)
#define copy_from_user(dst, src, len)	(memcpy(dst, src, len), 0)

void __i51_user_memset(char __user *buf, char c, user_size_t len);

/**
 * __user_memset - memset user memory, non-reentrant
 */
#define __user_memset(dst, val, len)	(__i51_user_memset(dst, val, len), 0)

#if 0
#if CONFIG_KERNEL_PREEMPT
void i51_user_memset(char __user *buf, char c, user_size_t len) __reentrant;

/**
 * user_memset - memset user memory, reentrant
 */
#define user_memset(dst, val, len)	(i51_user_memset(dst, val, len), 0)

#else
#define user_memset(dst, val, len)	__user_memset(dst, val, len)
#endif
#endif

void arch_init(void);

#ifdef BIG_ENDIAN
#define set16(dst, high, low) \
	do { \
		((uint8_t *)(dst))[0] = (high); \
		((uint8_t *)(dst))[1] = (low); \
	} while (0)
#else
#define set16(dst, high, low) \
	do { \
		((uint8_t *)(dst))[1] = (high); \
		((uint8_t *)(dst))[0] = (low); \
	} while (0)
#endif

#if CONFIG_PAGE_ALLOC
#define PAGE_SIZE	128
#define PAGE_NULL	(0xff)

#define NUM_PAGES	(CONFIG_XDATA_SIZE / PAGE_SIZE)

typedef unsigned char page_t;

page_t page_alloc(void) __reentrant;
void page_free(page_t) __reentrant;
#endif

#if CONFIG_TDATA
#define TDATA_BASE	0x80
#endif

#define put_user(x, ptr)	do { ((*ptr) = (x)); } while (0)
#define get_user(x, ptr)	do { ((x) = (*ptr)); } while (0)

#define __init_call(stage, x) \
	static void __ ## stage ## init_call(void) __naked \
	{ \
		__asm \
			.area stage ##_initcall(CODE,REL,CON) \
			.db _##x \
			.db _##x >> 8\
		__endasm; \
	}

#define arch_initcall(x) __init_call(arch, x)
#define subsys_initcall(x) __init_call(subsys, x)
#define device_initcall(x) __init_call(device, x)
#define late_initcall(x) __init_call(late, x)

#endif
